//
//	File:			Krislet.java
//	Author:		Krzysztof Langner
//	Date:			1997/04/28
//


import java.io.*;
import java.net.*;
import java.util.StringTokenizer;


//***************************************************************************
//
//	This is main object class
//
//***************************************************************************
class Krislet implements SendCommand
{

   public static final boolean DEBUG=false;

//===========================================================================
// Initialization member functions

	//---------------------------------------------------------------------------
	// The main appllication function.
	// Command line format:
	//
	// krislet [-parameter value]
	//
	// Parameters:
	//
	//	host (default "localhost")
	//		The host name can either be a machine name, such as "java.sun.com"
	//		or a string representing its IP address, such as "206.26.48.100."
	//
	//	port (default 6000)
	//		Port number for communication with server
	//
	//	team (default Kris)
	//		Team name. This name can not contain spaces.
	//
	//
	public static void main(String a[])
		throws SocketException, IOException
	{
	   String	hostName = new String("localhost");
	   int			port = 6000;
	   String	team = new String("jeeff");

      m_type=TYPE_FULL_COURT;

		try
		{
			// First look for parameters
			for( int c = 0 ; c < a.length ; c += 2 )
			{
				if( a[c].compareTo("-host") == 0 )
				{
					hostName = a[c+1];
				}
				else if( a[c].compareTo("-port") == 0 )
				{
					port = Integer.parseInt(a[c+1]);
				}
				else if( a[c].compareTo("-team") == 0 )
				{
					team = a[c+1];
				}
            else if( a[c].compareTo("-type") == 0 )
				{
					m_type = Integer.parseInt(a[c+1]);
				}
				else
				{
					throw new Exception();
				}
			}
		}
		catch(Exception e)
		{
			System.err.println("");
			System.err.println("USAGE: krislet [-parameter value]");
			System.err.println("");
			System.err.println("    Parameters  value           default");
			System.err.println("    --------------------------------------------");
			System.err.println("    host        host_name       localhost");
			System.err.println("    port        port_number     6000");
			System.err.println("    team        team_name       Kris");
         System.err.println("    type        type of player  0=Full Court");
         System.err.println("                                1=Left Back");
         System.err.println("                                2=Left Front");
         System.err.println("                                3=Right Back");
         System.err.println("                                4=Right Front");
         System.err.println("                               10=Center Back");
         System.err.println("                               11=Left Middle");
         System.err.println("                               12=Right Middle");
			System.err.println("");
			System.err.println("    Example:");
			System.err.println("      krislet -host www.host.com -port 6000 -team Poland");
			System.err.println("    or");
			System.err.println("      krislet -host 193.117.005.223");
			return;
		}

		Krislet player = new Krislet(InetAddress.getByName(hostName),
																	port, team);

      System.out.println("Playing as: "+m_type);
		// enter main loop
		player.mainLoop();
	}

	//---------------------------------------------------------------------------
	// This constructor opens socket for  connection with server
	public Krislet(InetAddress host, int port, String team)
		throws SocketException
	{
		m_socket = new DatagramSocket();
		m_host = host;
		m_port = port;
		m_team = team;
	}

	//---------------------------------------------------------------------------
	// This destructor closes socket to server
	public void finalize()
	{
		m_socket.close();
	}


//===========================================================================
// Protected member functions

	//---------------------------------------------------------------------------
	// This is main loop for player
	protected void mainLoop() throws IOException
	{
		byte[] buffer = new byte[MSG_SIZE];
		DatagramPacket packet = new DatagramPacket(buffer, MSG_SIZE);

		// first we need to initialize connection with server
		init();

		m_socket.receive(packet);
		parseInitCommand(new String(buffer,0));
		m_port = packet.getPort();

		// Now we should be connected to the server
		// and we know side, player number and play mode
		while( true )
			parseSensorInformation(receive());
	}


//===========================================================================
// Implementation of SendCommand Interface

	//---------------------------------------------------------------------------
	// This function sends move command to the server
	public void move(double x, double y)
	{
		send("(move " + Double.toString(x) + " " + Double.toString(y) + ")");
	}

	//---------------------------------------------------------------------------
	// This function sends turn command to the server
	public void turn(double moment)
	{
      moment = normalize_angle(moment);
      //debug("(turn " + Double.toString(moment) + ")");
      //int i_moment = (int)moment;
		send("(turn " + moment + ")");
      this.lastAngle=moment;
      bodyDir += moment;
      bodyDir=normalize_angle(bodyDir);
	}

   public void turn_to(double adir)
   {
      turn(adir - bodyDir);
   }

	//---------------------------------------------------------------------------
	// This function sends dash command to the server
	public void dash(double power)
	{
      //debug("(dash " + Double.toString(power) + ")");
		send("(dash " + Double.toString(power) + ")");
	}

   public void dashToDist(double distance)
   {
      double power = 70 * distance;
      if (power > 100) power=100;
      dash(power);
   }

	//---------------------------------------------------------------------------
	// This function sends kick command to the server
	public void kick(double power, double direction)
	{
		send("(kick " + Double.toString(power) + " " + Double.toString(direction) + ")");
	}

   public void kickToDist(double distance, double direction)
   {
      double power = 30 * distance;
      if (power > 100) power=100;
      kick(power, direction);
   }

	//---------------------------------------------------------------------------
	// This function sends say command to the server
	public void say(String message)
	{
		send("(say " + message + ")");
	}

  /* public void look_around()
   {
      m_brain.m_memory.waitForNewInfo();
      int i;
      for (i=0; i < 4 ; i++)
      {
	      turn(80) ;
	      m_brain.m_memory.waitForNewInfo();
     }

   }*/

	//---------------------------------------------------------------------------
	// This function sends chage_view command to the server
	public void changeView(String angle, String quality)
	{
		send("(change_view " + angle + " " + quality + ")");
	}

	//---------------------------------------------------------------------------
	// This function parses initial message from the server
	protected void parseInitCommand(String message)	throws IOException
	{
		StringTokenizer	tokenizer = new StringTokenizer(message,"() ");
		
		// We need init token
		if( tokenizer.nextToken().compareTo("init") != 0)
		{
			throw new IOException(message);
		}

		// initialize player's brain
		m_brain = new Brain(this, 
												m_team, 
												tokenizer.nextToken().charAt(0), 
												Integer.parseInt(tokenizer.nextToken()),
												tokenizer.nextToken());

	}


//===========================================================================
// Here comes collection of communication function
	//---------------------------------------------------------------------------
	// This function sends initialization command to the server
	private void init()
	{
		send("(init " + m_team + "(version 4.00))");
	}

	//---------------------------------------------------------------------------
	// This function parses sensor information
	private void parseSensorInformation(String message)
	{
		// First check kind of information		
		if( message.charAt(1) == 's' )
		{
			VisualInfo	info = new VisualInfo(message);
			info.parse();
			m_brain.see(info);
		}
		else if( message.charAt(1) == 'h' )
			parseHear(message);
	}


	//---------------------------------------------------------------------------
	// This function parses hear information
	private void parseHear(String message)
	{
		// get hear information
		StringTokenizer	tokenizer = new StringTokenizer(message,"() ");
		int	time;
		String sender;

      debug("---> HEAR: "+message);

		// skipp hear token
		tokenizer.nextToken();
		time = Integer.parseInt( tokenizer.nextToken() );
		sender = tokenizer.nextToken();
		if( sender.compareTo("referee") == 0 )
			m_brain.hear(time, tokenizer.nextToken());
		else if( sender.compareTo("self") != 0 )
			m_brain.hear(time, Integer.parseInt(sender), tokenizer.nextToken());
	}


	//---------------------------------------------------------------------------
	// This function sends via socket message to the server
	private void send(String message)
	{
      debug("Sending: T="+System.currentTimeMillis()+" msg="+message);
		byte[] buffer = new byte[MSG_SIZE];
		message.getBytes(0, message.length(), buffer, 0);

		DatagramPacket packet = new DatagramPacket(buffer, MSG_SIZE, m_host, m_port);

		try
      {
			m_socket.send(packet);
		}catch(IOException e){
			System.err.println("socket sending error " + e);
		}
	}

	//---------------------------------------------------------------------------
	// This function waits for new message from server
	private String receive()
	{
		byte[] buffer = new byte[MSG_SIZE];
		DatagramPacket packet = new DatagramPacket(buffer, MSG_SIZE);
		try{
			m_socket.receive(packet);
		}catch(IOException e){
			System.err.println("socket receiving error " + e);
		}

      String message = new String(buffer,0);
      message = message.trim();

      debug("RECEIVED:"+message.trim());
		//return new String(buffer,0);
      return (message);
      //return ("(see 0 ((goal r) 102.471 11) ((flag c) 52 22) ((flag c t) 50 -16) ((flag r t) 101.47 -7) ((flag r b) 114.089 28) ((flag p l t) 12.0011 0 0 0) ((flag pr t) 84.0002 0) ((flag p r c) 86.3481 13) ((flag p r b) 93.1065 25) ((flag g r t) 101.82 9) ((flag g r b) 103.247 13) ((flag t l 20) 33.8378 -34) ((flag t l 10) 42.4853 -26) ((flag t 0) 51.6236 -21) ((flag t r 10) 61.0328 -18) ((flag t r 20) 70.6045 -15) ((flag t r 30) 80.2808 -13) ((flag t r 40) 90.0278 -12) ((flag tr 50) 99.8248 -10) ((flag b r 20) 90.0278 40) ((flag b r 30) 97.8008 37) ((flagb r 40) 105.948 33) ((flag b r 50) 114.39 31) ((flag r t 30) 105.973 -5) ((flagr t 20) 105.5 0) ((flag r t 10) 105.973 5) ((flag r 0) 107.379 10) ((flag r b 10) 109.682 15) ((flag r b 20) 112.828 20) ((flag r b 30) 116.749 25) ((ball) 52 22) ((Player) 2 -179) ((line r) 100.5 -90))");
	}

   public double normalize_angle(double ang)
   {
      while(ang >  180.0) ang -= 360.0 ;
      while(ang < -180.0) ang += 360.0 ;
      return ang ;
   }

   public int getType()
   {
      return(m_type);
   }


   private void debug(String text)
   {
      if (DEBUG) System.out.println(text);
   }


//===========================================================================
// Private members
	// class memebers
	private DatagramSocket	m_socket;		// Socket to communicate with server
	private	InetAddress			m_host;			// Server address
	private int							m_port;			// server port
	public static String					m_team;			// team name

   public static int server_time=0;

	private SensorInput			m_brain;		// input for sensor information
   //private Brain			m_brain;		// input for sensor information

	// constants
	private static final int	MSG_SIZE = 1024;	// Size of socket buffer

   public static double bodyDir=0;

   //type of player
   private static int m_type;

   public static final int TYPE_FULL_COURT=0;
   public static final int TYPE_LEFT_BACK=1;
   public static final int TYPE_LEFT_FRONT=2;
   public static final int TYPE_RIGHT_BACK=3;
   public static final int TYPE_RIGHT_FRONT=4;

   public static final int TYPE_CENTER_BACK=10;
   public static final int TYPE_LEFT_MIDDLE=11;
   public static final int TYPE_RIGHT_MIDDLE=12;

   public static double lastAngle=0;


   //position of player in court
   public static Pos pos=new Pos();

}
